from typing import Dict, List, Any, Union, Optional, Callable, Tuple
from numbers import Number
import json
import warnings
import pandas as pd
import numpy as np
from macrosynergy.management.utils import (
    qdf_to_ticker_df,
    ticker_df_to_qdf,
    concat_single_metric_qdfs,
    get_cid,
    get_xcat,
    is_valid_iso_date,
)
from macrosynergy.management.types import QuantamentalDataFrame


def _get_diff_data(
    diff_mask: pd.Series,
    val_series: pd.Series,
    eop_series: pd.Series,
    grading_series: pd.Series,
    fvi: pd.Timestamp,
) -> pd.DataFrame:
    """
    Get the diff data for a given ticker from wide/pivoted DataFrames (`ticker_df`) of
    metrics `value`, `eop_lag` and `grading`.

    :param <pd.Series> diff_mask: A boolean mask indicating where the value has changed.
    :param <pd.Series> val_series: The value series.
    :param <pd.Series> eop_series: The end-of-period lag series.
    :param <pd.Series> grading_series: The grading series.
    :param <pd.Timestamp> fvi: The first valid index (fvi) for the ticker.
    :return: A DataFrame with the diff data.
    :rtype: pd.DataFrame
    """
    # get the first index as well
    dates = val_series.index[diff_mask].union([fvi])

    # create the diff dataframe
    df_temp: pd.DataFrame = pd.concat(
        (
            val_series.loc[dates].to_frame("value"),
            eop_series.loc[dates].to_frame("eop_lag"),
            grading_series.loc[dates].to_frame("grading"),
        ),
        axis=1,
        ignore_index=False,
    )

    df_temp["eop"] = df_temp.index - pd.to_timedelta(df_temp["eop_lag"], unit="D")
    df_temp["release"] = df_temp["eop_lag"].diff(periods=1) < 0

    df_temp = df_temp.sort_index().reset_index()
    df_temp["count"] = df_temp.index

    df_temp = pd.merge(
        left=df_temp,
        right=df_temp.groupby(["eop"], as_index=False)["count"].min(),
        on=["eop"],
        how="outer",
        suffixes=(None, "_min"),
    )
    df_temp["version"] = df_temp["count"] - df_temp["count_min"]
    df_temp["diff"] = df_temp["value"].diff(periods=1)
    df_temp = df_temp.set_index("real_date")[
        ["value", "eop", "version", "grading", "diff"]
    ]

    return df_temp


def _get_diff_density_stats(
    diff_mask: pd.Series, val_series: pd.Series, fvi: pd.Timestamp, lvi: pd.Timestamp
) -> Dict[str, Union[float, str]]:
    """
    Get the density stats for a given ticker from a boolean mask indicating where the value
    has changed and the value series.

    :param <pd.Series> diff_mask: A boolean mask indicating where the value has changed.
    :param <pd.Series> val_series: The value series.
    :return: A dictionary with the density stats.
    """
    diff_density: Number = 100 * diff_mask.sum() / (~val_series.isna()).sum()
    dtrange_str = f"{fvi.strftime('%Y-%m-%d')} : {lvi.strftime('%Y-%m-%d')}"
    return {"diff_density": diff_density, "date_range": dtrange_str}


def create_delta_data(
    df: QuantamentalDataFrame, return_density_stats: bool = False
) -> Union[Dict[str, pd.DataFrame], pd.DataFrame]:
    """
    Creates a dictionary of dataframes with the changes in the information state for each
    ticker in the QuantamentalDataFrame. Optionally, returns a DataFrame with the statistics
    for change frequency, density and date range for each ticker.
    :param <QuantamentalDataFrame> df: The QuantamentalDataFrame to calculate the changes for.
    :param <bool> return_density_stats: If True, returns a DataFrame with the density stats for each ticker.

    :return: A dictionary of DataFrames with the changes in the information state for each ticker.
    """

    if not isinstance(df, QuantamentalDataFrame):
        raise ValueError("`df` must be a QuantamentalDataFrame")
    if not isinstance(return_density_stats, bool):
        raise ValueError("`return_density_stats` must be a boolean")
    if "value" not in df.columns:
        raise ValueError("`df` must contain a `value` column")
    if "eop_lag" not in df.columns:
        df["eop_lag"] = np.nan
    if "grading" not in df.columns:
        df["grading"] = np.nan

    values_df = qdf_to_ticker_df(df, value_column="value")
    eop_df = qdf_to_ticker_df(df, value_column="eop_lag")
    grading_df = qdf_to_ticker_df(df, value_column="grading")
    assert set(values_df.columns) == set(eop_df.columns) == set(grading_df.columns)
    all_tickers: List[str] = values_df.columns.tolist()

    # get the first valid index for each column
    fvi_series: pd.Series = values_df.apply(lambda x: x.first_valid_index())
    lvi_series: pd.Series = values_df.apply(lambda x: x.last_valid_index())

    # create dicts to store the dataframes and density stats
    isc_dict: Dict[str, Any] = {}
    # density_stats: Dict[str, Dict[str, Any]] = {}

    diff_mask = values_df.diff(axis=0).abs() > 1e-12
    density_stats: Dict[str, Dict[str, Union[float, str]]] = {}

    for ticker in all_tickers:
        isc_dict[ticker] = _get_diff_data(
            diff_mask=diff_mask[ticker],
            val_series=values_df[ticker],
            eop_series=eop_df[ticker],
            grading_series=grading_df[ticker],
            fvi=fvi_series[ticker],
        )
        density_stats[ticker] = _get_diff_density_stats(
            diff_mask=diff_mask[ticker],
            val_series=values_df[ticker],
            fvi=fvi_series[ticker],
            lvi=lvi_series[ticker],
        )

    # flatten the density stats
    _dstats_flat = [
        (k, v["diff_density"], v["date_range"]) for k, v in density_stats.items()
    ]

    density_stats_df = (
        pd.DataFrame(_dstats_flat, columns=["ticker", "changes_density", "date_range"])
        .sort_values(by="changes_density", ascending=False)
        .reset_index(drop=True)
    )
    if return_density_stats:
        return isc_dict, density_stats_df
    else:
        return isc_dict


class SubscriptableMeta(type):
    """
    Convenience metaclass to allow subscripting of methods on a class.
    """

    def __getitem__(cls, item):
        if hasattr(cls, item) and callable(getattr(cls, item)):
            return getattr(cls, item)
        else:
            raise KeyError(f"{item} is not a valid method name")


class VolatilityEstimationMethods(metaclass=SubscriptableMeta):
    """
    Class to hold methods for calculating standard deviations.
    Each method must comply to the following signature:
        `func(s: pd.Series, **kwargs) -> pd.Series`

    Currently supported methods are:

    - `std`: Standard deviation
    - `abs`: Mean absolute deviation
    - `exp`: Exponentially weighted standard deviation
    - `exp_abs`: Exponentially weighted mean absolute deviation
    """

    @staticmethod
    def std(s: pd.Series, min_periods: int, **kwargs) -> pd.Series:
        """
        Calculate the expanding standard deviation of a Series.

        :param <pd.Series> s: The Series to calculate the standard deviation for.
        :param <int> min_periods: The minimum number of periods required for the calculation.
        :return: The standard deviation of the Series.
        """
        return s.expanding(min_periods=min_periods).std()

    @staticmethod
    def abs(s: pd.Series, min_periods: int, **kwargs) -> pd.Series:
        """
        Calculate the expanding mean absolute deviation of a Series.

        :param <pd.Series> s: The Series to calculate the absolute standard deviation for.
        :param <int> min_periods: The minimum number of periods required for the calculation.
        :return: The absolute standard deviation of the Series.
        """
        mean = s.expanding(min_periods=min_periods).mean()
        return (s - mean.bfill()).abs().expanding(min_periods=min_periods).mean()

    @staticmethod
    def exp(s: pd.Series, halflife: int, min_periods: int, **kwargs) -> pd.Series:
        """
        Calculate the exponentially weighted standard deviation of a Series.

        :param <pd.Series> s: The Series to calculate the exponentially weighted standard
            deviation for.
        :param <int> halflife: The halflife of the exponential weighting.
        :param <int> min_periods: The minimum number of periods required for the
            calculation.
        :return: The exponentially weighted standard deviation of the Series.
        """
        return s.ewm(halflife=halflife, min_periods=min_periods).std()

    @staticmethod
    def exp_abs(s: pd.Series, halflife: int, min_periods: int, **kwargs) -> pd.Series:
        """
        Calculate the exponentially weighted mean absolute deviation of a Series.

        :param <pd.Series> s: The Series to calculate the exponentially weighted absolute
            standard deviation for.
        :param <int> halflife: The halflife of the exponential weighting.
        :param <int> min_periods: The minimum number of periods required for the calculation.
        :return: The exponentially weighted absolute standard deviation of the Series.
        """
        mean = s.ewm(halflife=halflife, min_periods=min_periods).mean()
        sd = (
            (s - mean.bfill())
            .abs()
            .ewm(halflife=halflife, min_periods=min_periods)
            .mean()
        )
        return sd


CALC_SCORE_CUSTOM_METHOD_ERR_MSG = (
    "Method {std} not supported. "
    f"Supported methods are: {dir(VolatilityEstimationMethods)}. \n"
    "Alternatively, provide a custom method with signature "
    "`custom_method(s: pd.Series, **kwargs) -> pd.Series` "
    "using the `custom_method` and `custom_method_kwargs` arguments."
)


def calculate_score_on_sparse_indicator(
    isc: Dict[str, pd.DataFrame],
    std: str = "std",
    halflife: int = None,
    min_periods: int = 10,
    isc_version: int = 0,
    iis: bool = False,
    custom_method: Optional[Callable] = None,
    custom_method_kwargs: Dict = {},
    volatility_forecast: bool = True,
) -> Dict[str, pd.DataFrame]:
    """Calculate score on sparse indicator

    :param <Dict[str, pd.DataFrame]> isc: A dictionary of DataFrames with the changes in
        the information state for each ticker.

    :param <str> std: The method to use for calculating the standard deviation.
        Supported methods are `std`, `abs`, `exp` and `exp_abs`. See the documentation for
        `VolatilityEstimationMethods` for more information.
    :param <int> halflife: The halflife of the exponential weighting. Only used with `exp`
        and `exp_abs` methods. Default is None.
    :param <int> min_periods: The minimum number of periods required for the calculation.
        Default is 10.
    :param <int> isc_version: The version of the information state changes to use. If set
        to 0 (default), only the first version is used. If set to any other positive integer,
        all versions are used.
    :param <bool> iis: if True (default) zn-scores are also calculated for the initial
        sample period defined by `min_periods`, on an in-sample basis, to avoid losing history.

    :param <Callable> custom_method: A custom method to use for calculating the standard
        deviation. Must have the signature `custom_method(s: pd.Series, **kwargs) -> pd.Series`.
    :param <Dict> custom_method_kwargs: Keyword arguments to pass to the custom method.
    :param <bool> volatility_forecast: If True (default), the volatility forecast is shifted
        one period forward to align with the information state changes.

    :return: A dictionary of DataFrames with the changes in the information state for each ticker.
    """
    # Operations on a per key in data dictionary

    curr_method: Callable[[pd.Series, Optional[Dict[str, Any]]], pd.Series]
    if custom_method is not None:
        if not callable(custom_method):
            raise TypeError("`custom_method` must be a callable")
        if not isinstance(custom_method_kwargs, dict):
            raise TypeError("`custom_method_kwargs` must be a dictionary")
        curr_method = custom_method
    else:
        if not hasattr(VolatilityEstimationMethods, std):
            raise ValueError(CALC_SCORE_CUSTOM_METHOD_ERR_MSG.format(std=std))
        # curr_method = getattr(StandardDeviationMethod, std)
        curr_method = VolatilityEstimationMethods[std]

    method_kwargs: Dict[str, Any] = dict(
        min_periods=min_periods, halflife=halflife, **custom_method_kwargs
    )
    # if not 0, then use all versions
    for key, v in isc.items():
        mask_rel = (v["version"] == 0) if isc_version == 0 else (v["version"] >= 0)
        s = v.loc[mask_rel, "diff"]
        # TODO exponential weights (requires knowledge of frequency...)

        result: pd.Series = curr_method(s, **method_kwargs)

        columns = [kk for kk in v.columns if kk != "std"]
        v = pd.merge(
            left=v[columns],
            right=result.to_frame("std").shift(periods=int(volatility_forecast)),
            how="left",
            left_index=True,
            right_index=True,
        )
        v["std"] = v["std"].ffill()
        if iis:
            v["std"] = v["std"].bfill()
        v["zscore"] = v["diff"] / v["std"]

        isc[key] = v

    return isc


def _infer_frequency_timeseries(eop_lag_series: pd.Series) -> Optional[str]:
    """
    Infer the frequency of a time series based on the pattern of eop_lag values,
    considering resets to 0 to detect period ends.
    :param <pd.Series> eop_lag_series: A Series of eop_lag values.
    :return: The inferred frequency of the time series. One of "D", "W", "M", "Q" or "A".
    """
    # Identify resets to 0 which indicate the end of a period
    reset_indices = eop_lag_series[eop_lag_series == 0].index
    if len(reset_indices) < 2:
        return "D"  # Default to daily if insufficient data

    # Calculate periods between resets
    periods = reset_indices.to_series().diff().dropna().dt.days

    # Determine the most common period length
    most_common_period = periods.mode().values[0]

    # Define ranges for different frequencies
    freqs = {"D": 1, "W": 7, "M": 30, "Q": 91, "A": 365}
    # 10% tolerance for frequency ranges
    for freq, freq_range in freqs.items():
        rng = (1 - 0.1) * freq_range, (1 + 0.1) * freq_range
        if rng[0] <= most_common_period <= rng[1]:
            return freq

    return "D"  # Default to daily if no match is found


def infer_frequency(df: QuantamentalDataFrame) -> pd.Series:
    """
    Infer the frequency of a QuantamentalDataFrame based on the most common eop_lag values.
    :param <QuantamentalDataFrame> df: The QuantamentalDataFrame to infer the frequency for.
    :return: A Series with the inferred frequency for each ticker in the QuantamentalDataFrame.
    """
    if not isinstance(df, QuantamentalDataFrame):
        raise TypeError("`df` must be a QuantamentalDataFrame")
    if not "eop_lag" in df.columns:
        raise ValueError("`df` must contain an `eop_lag` column")

    ticker_df = qdf_to_ticker_df(df, value_column="eop_lag")
    freq_dict = {
        ticker: _infer_frequency_timeseries(ticker_df[ticker])
        for ticker in ticker_df.columns
    }
    return pd.Series(freq_dict)


def weight_from_frequency(freq: str, base: float = 252):
    """Weight from frequency"""
    # TODO apply on multiple tickers
    freq_map = {"D": 1, "W": 5, "M": 21, "Q": 93, "A": 252}
    assert freq in freq_map, f"Frequency {freq} not supported"
    return freq_map[freq] / base


def _remove_insignificant_values(
    df: pd.DataFrame, threshold: float = 1e-12
) -> pd.DataFrame:
    """
    Convenience function to remove insignificant values from a DataFrame.

    :param <pd.DataFrame> df: The DataFrame to remove insignificant values from.
    :param <float> threshold: The threshold below which values are considered insignificant.
    :return: The DataFrame with insignificant values removed.
    """
    return df / (df.cumsum(axis=0).abs() > threshold).astype(int)


def _isc_dict_to_frames(
    isc: Dict[str, pd.DataFrame], metric: str = "value"
) -> List[pd.DataFrame]:
    """
    Convert a dictionary of DataFrames to a list of DataFrames with a specific metric.
    """
    frames = []
    for k, v in isc.items():
        assert isinstance(v, pd.DataFrame)
        assert metric in v.columns
        _fr = v[metric].to_frame(k)
        if _fr.empty:
            warnings.warn(f"Empty frame for {k}")
            continue

        frames.append(_fr)

    return frames


def _get_metric_df_from_isc(
    isc: Dict[str, pd.DataFrame],
    metric: str,
    date_range: pd.DatetimeIndex,
    fill: Union[str, Number] = 0,
) -> pd.DataFrame:
    """
    Get a DataFrame with a specific metric from a dictionary of DataFrames.

    :param <Dict[str, pd.DataFrame]> isc: A dictionary of DataFrames with the changes in
        the information state for each ticker.
    :param <str> metric: The name of the metric to extract.
    :param <pd.DatetimeIndex> date_range: The date range to reindex the DataFrame to.
    :param <Union[str, Number]> fill: The value to fill NaNs with. If 'ffill', forward fill
        NaNs. Default is 0.
    """
    fill_err: str = "`fill` must be a number to replace NaNs or 'ffill' to forward fill"
    if not isinstance(fill, (str, Number)):
        raise TypeError(fill_err)
    if isinstance(fill, str):
        if not fill == "ffill":
            raise ValueError(fill_err)

    df_frames: List[pd.DataFrame] = _isc_dict_to_frames(isc, metric=metric)
    tdf: pd.DataFrame = pd.concat(df_frames, axis=1).reindex(date_range)
    if fill == "ffill":
        tdf = tdf.ffill()
    else:
        tdf = tdf.fillna(fill)
    tdf.columns.name, tdf.index.name = "ticker", "real_date"
    return tdf


def sparse_to_dense(
    isc: Dict[str, pd.DataFrame],
    value_column: str,
    min_period: pd.Timestamp,
    max_period: pd.Timestamp,
    postfix: str = None,
    metrics: List[str] = ["eop", "grading"],
) -> pd.DataFrame:
    """
    Convert a dictionary of DataFrames with changes in the information state to a dense
    DataFrame (QuantamentalDataFrame).

    :param <Dict[str, pd.DataFrame]> isc: A dictionary of DataFrames with the changes in
        the information state for each ticker.
    :param <str> value_column: The name of the column to use as the value.
    :param <pd.Timestamp> min_period: The minimum period to include in the DataFrame.
    :param <pd.Timestamp> max_period: The maximum period to include in the DataFrame.
    :param <str> postfix: A postfix to append to the xcat column. Default is None.
    :param <List[str]> metrics: A list of metrics to include in the DataFrame. Default is
        ["eop", "grading"].
    :return: A dense DataFrame with the changes in the information state.
    """

    # TODO store real_date min and max in object...
    dtrange = pd.date_range(
        start=min_period,
        end=max_period,
        freq="B",
        inclusive="both",
    )

    tdf = _get_metric_df_from_isc(isc=isc, metric=value_column, date_range=dtrange)
    tdf = _remove_insignificant_values(tdf, threshold=1e-12)

    sm_qdfs: List[QuantamentalDataFrame] = [ticker_df_to_qdf(tdf)]
    for metric_name in metrics:
        wdf = _get_metric_df_from_isc(
            isc=isc, metric=metric_name, date_range=dtrange, fill="ffill"
        )
        if wdf.empty or wdf.isna().all().all():
            dfs = [
                pd.DataFrame(index=dtrange)
                .assign(
                    cid=get_cid(tickerx),
                    xcat=get_xcat(tickerx),
                    obs=np.nan,
                )
                .rename(columns={"obs": metric_name})
                .reset_index()
                for tickerx in wdf.columns
            ]
            m_qdf = pd.concat(dfs, axis=0).reset_index(drop=True)
        else:
            m_qdf = ticker_df_to_qdf(wdf, metric=metric_name)

        sm_qdfs.append(m_qdf)

    qdf: QuantamentalDataFrame = concat_single_metric_qdfs(sm_qdfs)
    if "eop" in metrics:
        qdf["eop_lag"] = (qdf["real_date"] - qdf["eop"]).dt.days

    if postfix:
        qdf["xcat"] += postfix

    return qdf


def temporal_aggregator_exponential(
    df: QuantamentalDataFrame,
    halflife: int = 5,
    winsorise: float = None,
) -> pd.DataFrame:
    """
    Temporal aggregator using exponential moving average.

    :param <QuantamentalDataFrame> df: The QuantamentalDataFrame to aggregate.
    :param <int> halflife: The halflife of the exponential moving average.
    :param <float> winsorise: The value to winsorise the data to. Default is None.
    :return: A DataFrame with the aggregated values.
    """
    tdf: pd.DataFrame = qdf_to_ticker_df(df)
    if winsorise:
        tdf = tdf.clip(lower=-winsorise, upper=winsorise)
    # Exponential moving average weights
    tdf = tdf.ewm(halflife=halflife).mean()
    qdf: QuantamentalDataFrame = ticker_df_to_qdf(tdf)
    qdf["xcat"] += f"EWM{halflife:d}D"
    return qdf


def temporal_aggregator_mean(
    df: QuantamentalDataFrame,
    window: int = 21,
    winsorise: float = None,
) -> pd.DataFrame:
    """
    Temporal aggregator using a rolling mean.

    :param <QuantamentalDataFrame> df: The QuantamentalDataFrame to aggregate.
    :param <int> window: The window size for the rolling mean.
    :param <float> winsorise: The value to winsorise the data to. Default is None.
    :return: A DataFrame with the aggregated values.
    """
    tdf: pd.DataFrame = qdf_to_ticker_df(df)
    if winsorise:
        tdf = tdf.clip(lower=-winsorise, upper=winsorise)

    tdf = tdf.rolling(window=window).mean()
    qdf: QuantamentalDataFrame = ticker_df_to_qdf(tdf)
    qdf["xcat"] += f"MA{window:d}D"
    return qdf


def temporal_aggregator_period(
    isc: Dict[str, pd.DataFrame],
    start: pd.Timestamp,
    end: pd.Timestamp,
    winsorise: int = 10,
    postfix: str = "_NCSUM",
) -> pd.DataFrame:
    """Temporal aggregator over periods of changes in the information state.

    :param <Dict[str, pd.DataFrame]> isc: A dictionary of DataFrames with the changes in
        the information state for each ticker.
    :param <pd.Timestamp> start: The start date of the period to aggregate.
    :param <pd.Timestamp> end: The end date of the period to aggregate.
    :param <int> winsorise: The value to winsorise the data to. Default is 10.
    :param <str> postfix: A postfix to append to the xcat column. Default is "_NCSUM".
    :return: A DataFrame with the aggregated values.
    """
    dt_range = pd.date_range(start=start, end=end, freq="B", inclusive="both")
    tdf: pd.DataFrame = _get_metric_df_from_isc(
        isc=isc,
        metric="zscore_norm_squared",
        date_range=dt_range,
        fill=0,
    )
    # Winsorise and remove insignificant values
    tdf = _remove_insignificant_values(tdf, threshold=1e-12)
    tdf = tdf.clip(lower=-winsorise, upper=winsorise)

    # Map out the eop dates
    p_eop: pd.DataFrame = _get_metric_df_from_isc(
        isc=isc,
        metric="eop",
        date_range=dt_range,
        fill="ffill",
    )

    qdf: QuantamentalDataFrame = concat_single_metric_qdfs(
        [
            ticker_df_to_qdf(tdf),
            ticker_df_to_qdf(p_eop, metric="eop"),
        ]
    )

    # group by cid, xcat, eop
    qdf["value"] = (
        qdf.groupby(["cid", "xcat", "eop"])["value"].cumsum().reset_index(drop=True)
    )
    if postfix:
        qdf["xcat"] += postfix
    return qdf


def _calculate_score_on_sparse_indicator_for_class(
    cls: "InformationStateChanges",
    std: str = "std",
    halflife: int = None,
    min_periods: int = 10,
    isc_version: int = 0,
    iis: bool = False,
    custom_method: Optional[Callable] = None,
    custom_method_kwargs: Dict = {},
    volatility_forecast: bool = True,
):
    """
    Calculate score on sparse indicator for a class.
    Effectively a re-implementation of the function `calculate_score_on_sparse_indicator`
    that specifically operates on an `InformationStateChanges` object.
    """
    assert isinstance(
        cls, InformationStateChanges
    ), "`cls` must be an `InformationStateChanges` object"
    assert hasattr(cls, "isc_dict") and isinstance(
        cls.isc_dict, dict
    ), "`InformationStateChanges` object not initialized"

    curr_method: Callable[[pd.Series, Optional[Dict[str, Any]]], pd.Series]
    if custom_method is not None:
        if not callable(custom_method):
            raise TypeError("`custom_method` must be a callable")
        if not isinstance(custom_method_kwargs, dict):
            raise TypeError("`custom_method_kwargs` must be a dictionary")
        curr_method = custom_method
    else:
        if not hasattr(VolatilityEstimationMethods, std):
            raise ValueError(CALC_SCORE_CUSTOM_METHOD_ERR_MSG.format(std=std))
        curr_method = VolatilityEstimationMethods[std]

    method_kwargs: Dict[str, Any] = dict(
        min_periods=min_periods, halflife=halflife, **custom_method_kwargs
    )
    for key, v in cls.isc_dict.items():
        mask_rel = (v["version"] == 0) if isc_version == 0 else (v["version"] >= 0)
        s = v.loc[mask_rel, "diff"]
        result: pd.Series = curr_method(s, **method_kwargs)
        columns = [kk for kk in v.columns if kk != "std"]
        v = pd.merge(
            left=v[columns],
            right=result.to_frame("std").shift(periods=int(volatility_forecast)),
            how="left",
            left_index=True,
            right_index=True,
        )
        v["std"] = v["std"].ffill()
        if iis:
            v["std"] = v["std"].bfill()
        v["zscore"] = v["diff"] / v["std"]

        cls.isc_dict[key] = v


class InformationStateChanges(object):
    """
    Class to hold information state changes for a set of tickers.

    Initialize using the `from_qdf` class method to create an `InformationStateChanges`
    object from a `QuantamentalDataFrame`. The `calculate_score` method can be used to
    calculate scores for the information state changes.
    """

    def __init__(
        self,
        min_period: pd.Timestamp = None,
        max_period: pd.Timestamp = None,
    ):
        self.isc_dict: Dict[str, pd.DataFrame] = dict()
        self.density_stats_df: pd.DataFrame = None
        self._min_period: pd.Timestamp = min_period
        self._max_period: pd.Timestamp = max_period

    def __getitem__(self, item):
        return self.isc_dict[item]

    def __setitem__(self, key, value):
        self.isc_dict[key] = value

    def __str__(self):
        return str(self.isc_dict)

    def __repr__(self):
        return repr(self.isc_dict)

    def keys(self):
        """
        A list of tickers in the InformationStateChanges object.
        """
        return self.isc_dict.keys()

    def values(self):
        """
        Extract the DataFrames from the InformationStateChanges object.
        """
        return self.isc_dict.values()

    def items(self):
        """
        Iterate through (ticker, DataFrame) pairs in the InformationStateChanges object.
        """
        return self.isc_dict.items()

    @classmethod
    def from_qdf(
        cls, qdf: QuantamentalDataFrame, norm: bool = True, **kwargs
    ) -> "InformationStateChanges":
        """
        Create an InformationStateChanges object from a QuantamentalDataFrame.

        :param <QuantamentalDataFrame> qdf: The QuantamentalDataFrame to create the
            InformationStateChanges object from.
        :param <bool> norm: If True, calculate the score for the information state changes.
        :param <**kwargs>: Additional keyword arguments to pass to the `calculate_score`
            method.
        :return: An InformationStateChanges object.
        """

        isc = cls(min_period=qdf["real_date"].min(), max_period=qdf["real_date"].max())
        isc_dict, density_stats_df = create_delta_data(qdf, return_density_stats=True)

        isc.isc_dict = isc_dict
        isc.density_stats_df = density_stats_df

        if norm:
            isc.calculate_score(**kwargs)

        return isc

    def to_qdf(
        self,
        value_column: str = "value",
        postfix: str = None,
        metrics: List[str] = ["eop", "grading"],
    ) -> pd.DataFrame:
        """
        Convert the InformationStateChanges object to a QuantamentalDataFrame.

        :param <str> value_column: The name of the column to use as the value.
        :param <str> postfix: A postfix to append to the xcat column. Default is None.
        :param <List[str]> metrics: A list of metrics to include in the DataFrame. Default is
            ["eop", "grading"].
        :return: A QuantamentalDataFrame with the information state changes.
        """
        return sparse_to_dense(
            isc=self.isc_dict,
            value_column=value_column,
            min_period=self._min_period,
            max_period=self._max_period,
            postfix=postfix,
            metrics=metrics,
        )

    def to_dict(
        self, ticker: str
    ) -> Dict[
        str, Union[List[Tuple[str, float, str, float]], Tuple[str, str, str], str]
    ]:
        data = [
            (f"{index:%Y-%m-%d}", row.value, f"{row.eop:%Y-%m-%d}", row.grading)
            for index, row in self[ticker][["value", "eop", "grading"]].iterrows()
        ]

        columns = ("real_date", "value", "eop", "grading")
        return_dict = {
            "data": data,
            "columns": columns,
            "last_real_date": f"{self._max_period:%Y-%m-%d}",
            "ticker": ticker,
        }
        return return_dict

    def to_json(self, ticker: str) -> str:
        return json.dumps(self.to_dict(ticker))

    def get_releases(
        self,
        from_date: Optional[Union[pd.Timestamp, str]] = pd.Timestamp.today().normalize()
        - pd.offsets.BDay(1),
        to_date: Optional[Union[pd.Timestamp, str]] = pd.Timestamp.today().normalize(),
        excl_xcats: List[str] = None,
        latest_only: bool = True,
    ) -> pd.DataFrame:
        """
        Get the latest releases for the InformationStateChanges object.

        :param <pd.Timestamp> from_date: The start date of the period to get releases for.
        :param <pd.Timestamp> to_date: The end date of the period to get releases for.
        :param <List[str]> excl_xcats: A list of xcats to exclude from the releases.
        :param <bool> latest_only: If True, only the latest release for each ticker is
            returned. Default is True.
        :return: A DataFrame with the latest releases for each ticker. If `latest_only` is
            False, all releases within the date range are returned.
        """

        if excl_xcats is not None:
            excl_xcat_err = "`excl_xcats` must be a list of strings"
            if not isinstance(excl_xcats, list):
                raise TypeError(excl_xcat_err)
            if not all(isinstance(x, str) for x in excl_xcats):
                raise TypeError(excl_xcat_err)
        else:
            excl_xcats = []

        if not isinstance(latest_only, bool):
            raise ValueError("`latest_only` must be a boolean")

        dt_err = "`{varname}` must be a `pd.Timestamp` or an ISO formatted date"
        for var_name in ["from_date", "to_date"]:
            if not isinstance(eval(var_name), (pd.Timestamp, str, type(None))):
                raise TypeError(dt_err.format(varname=var_name))
            if isinstance(eval(var_name), str):
                is_valid_iso_date(eval(var_name))

        if from_date is None:
            from_date = self._min_period
        elif isinstance(from_date, str):
            from_date = pd.Timestamp(from_date)
        if to_date is None:
            to_date = self._max_period
        elif isinstance(to_date, str):
            to_date = pd.Timestamp(to_date)

        if from_date > to_date:
            from_date, to_date = to_date, from_date
            warnings.warn("`from_date` is greater than `to_date`. Swapping the dates.")

        dfs_list = []
        for k, v in self.items():
            if get_xcat(k) in excl_xcats:
                continue
            s: pd.DataFrame = v.copy()
            s = s[(s.index >= from_date) & (s.index <= to_date)]
            s["ticker"] = k
            if latest_only and not s.empty:
                s = s.loc[[s.last_valid_index()]]

            dfs_list.append(s.reset_index())

        rel = (
            pd.concat(dfs_list, axis=0)
            .sort_values(by=["real_date", "eop", "ticker"])
            .rename(columns={"diff": "change"})
            .reset_index(drop=True)
        )

        if latest_only:
            return rel.set_index("ticker")
        else:
            return rel

    def temporal_aggregator_period(
        self,
        winsorise: int = 10,
        start: Optional[pd.Timestamp] = None,
        end: Optional[pd.Timestamp] = None,
    ) -> pd.DataFrame:
        """
        Temporal aggregator over periods of changes in the information state.

        :param <int> winsorise: The value to winsorise the data to. Default is 10.
        :param <pd.Timestamp> start: The start date of the period to aggregate.
        :param <pd.Timestamp> end: The end date of the period to aggregate.
        :return: A DataFrame with the aggregated values.
        """
        return temporal_aggregator_period(
            isc=self.isc_dict,
            start=start or self._min_period,
            end=end or self._max_period,
            winsorise=winsorise,
        )

    def calculate_score(
        self,
        std: str = "std",
        halflife: int = None,
        min_periods: int = 10,
        isc_version: int = 0,
        iis: bool = False,
        custom_method: Optional[Callable] = None,
        custom_method_kwargs: Dict = {},
        volatility_forecast: bool = True,
    ):
        """
        Calculate score on sparse indicator for the InformationStateChanges object.

        :param <str> std: The method to use for calculating the standard deviation.
            Supported methods are `std`, `abs`, `exp` and `exp_abs`. See the documentation for
            `StandardDeviationMethods` for more information.

        :param <int> halflife: The halflife of the exponential weighting. Only used with `exp`
            and `exp_abs` methods. Default is None.
        :param <int> min_periods: The minimum number of periods required for the calculation.
            Default is 10.
        :param <int> isc_version: The version of the information state changes to use. If set
            to 0 (default), only the first version is used. If set to any other positive integer,
            all versions are used.
        :param <bool> iis: if True (default) zn-scores are also calculated for the initial
            sample period defined by `min_periods`, on an in-sample basis, to avoid losing history.
        :param <Callable> custom_method: A custom method to use for calculating the standard
            deviation. Must have the signature `custom_method(s: pd.Series, **kwargs) -> pd.Series`.
        :param <Dict> custom_method_kwargs: Keyword arguments to pass to the custom method.
        :param <bool> volatility_forecast: If True (default), the volatility forecast is shifted
            one period forward to align with the information state changes.
        :return: None
        """

        _calculate_score_on_sparse_indicator_for_class(
            cls=self,
            std=std,
            halflife=halflife,
            min_periods=min_periods,
            isc_version=isc_version,
            iis=iis,
            custom_method=custom_method,
            custom_method_kwargs=custom_method_kwargs,
            volatility_forecast=volatility_forecast,
        )
        return self



"""
Functions for calculating the hedge ratios of a panel of returns with respect to a
single return. 
"""

import warnings
import numpy as np
import pandas as pd
from typing import List, Tuple
import statsmodels.api as sm
from statsmodels.regression.linear_model import RegressionResults

from macrosynergy.management.types import QuantamentalDataFrame
from macrosynergy.management.simulate import make_qdf
from macrosynergy.management.utils import (
    reduce_df,
    _map_to_business_day_frequency,
    standardise_dataframe,
)
import matplotlib.pyplot as plt


def date_alignment(
    unhedged_return: pd.Series, benchmark_return: pd.Series
) -> Tuple[pd.Timestamp, pd.Timestamp]:
    """
    Method used to align the two Series over the same timestamps: the sample data for the
    endogenous & exogenous variables must match throughout the re-estimation calculation.

    :param <pd.DataFrame> unhedged_return: the return series of the asset that is being
        hedged.
    :param <pd.Series> benchmark_return: the return series of the asset being used to
        hedge against the main asset.

    :return <pd.Timestamp, pd.Timestamp>: the shared start and end date across the two
        series.
    """
    ma_dates = unhedged_return.index
    ha_dates = benchmark_return.index

    if ma_dates[0] > ha_dates[0]:
        start_date = ma_dates[0]
    else:
        start_date = ha_dates[0]

    if ma_dates[-1] > ha_dates[-1]:
        end_date = ha_dates[-1]
    else:
        end_date = ma_dates[-1]

    return start_date, end_date


def hedge_calculator(
    unhedged_return: pd.Series,
    benchmark_return: pd.Series,
    rdates: List[pd.Timestamp],
    cross_section: str,
    meth: str = "ols",
    min_obs: int = 24,
    max_obs: int = 1000,
) -> pd.DataFrame:
    """
    Calculate the hedge ratios for each cross-section in the panel being hedged. It is
    worth noting that the sample of data used for calculating the hedge ratio will
    increase according to the dates parameter: each date represents an additional number
    of timestamps where the numeracy of dates added to the sample is instructed by the
    "refreq" parameter.

    :param <pd.Series> unhedged_return: the return series of the asset that is being
        hedged.
    :param <pd.Series> benchmark_return: the return series of the asset being used to
        hedge against the main asset.
    :param <List[pd.Timestamp]> rdates: the dates controlling the frequency of
        re-estimation.
    :param <str> cross_section: cross-section responsible for the "benchmark_return"
        series.
    :param <str> meth: method to estimate hedge ratio. At present the only method is
        OLS regression ('ols').
    :param <int> min_obs: a hedge ratio will only be computed if the number of days has
        surpassed the integer held by the parameter.
    :param <int> max_obs: the maximum number of latest observations allowed in order to
        estimate a hedge ratio. The default value is 1000.

    :return <pd.DataFrame>: returns a dataframe of the hedge ratios for the respective
        cross-section.
    """

    benchmark_return = benchmark_return.astype(dtype=np.float32)
    unhedged_return = unhedged_return.astype(dtype=np.float32)

    if benchmark_return.name is None:
        benchmark_return.name = "hedge"

    benchmark_return = benchmark_return[
        benchmark_return.first_valid_index() : benchmark_return.last_valid_index()
    ]
    unhedged_return = unhedged_return[
        unhedged_return.first_valid_index() : unhedged_return.last_valid_index()
    ]

    s_date, e_date = date_alignment(
        unhedged_return=unhedged_return, benchmark_return=benchmark_return
    )

    unhedged_return = unhedged_return.truncate(before=s_date, after=e_date)
    benchmark_return = benchmark_return.truncate(before=s_date, after=e_date)

    # The date series will be adjusted to each cross-section. Daily dates each return
    # series is defined over.
    date_series = unhedged_return.index
    df_ur = unhedged_return.to_frame(name="returns")
    df_ur = df_ur.reset_index()

    # Access the minimum date from the adjusted series: having aligned the unhedged asset
    # and the benchmark return. Both series will be defined over the same timestamps.
    min_obs_date = date_series[min_obs]

    # Storage dataframe defined over the re-balancing dates.
    data_column = np.empty(len(rdates))
    data_column[:] = np.nan
    df_hrat = pd.DataFrame(data=data_column, index=rdates, columns=["value"])

    for d in rdates:
        if d > min_obs_date:
            curr_start_date: pd.Timestamp = rdates[max(0, rdates.index(d) - max_obs)]
            # Inclusive of the re-estimation date.
            yvar = unhedged_return.loc[curr_start_date:d]
            xvar = benchmark_return.loc[curr_start_date:d]
            # Condition currently redundant but will become relevant.
            if meth == "ols":
                xvar = sm.add_constant(xvar)
                results: RegressionResults = sm.OLS(yvar, xvar).fit()
                results_params: pd.Series = results.params

            df_hrat.loc[d] = results_params.loc[benchmark_return.name]

    # Any dates prior to the minimum observation which would be classified by NaN values
    # remove from the DataFrame.
    df_hrat = df_hrat.dropna(axis=0, how="all")
    df_hrat.index.name = "real_date"
    df_hrat = df_hrat.reset_index(level=0)

    # Merge to convert to the re-estimation frequency. The intermediary dates, daily
    # business days between re-estimation dates, will be populated with np.NaN values.
    df_hr = df_ur.merge(df_hrat, on="real_date", how="left")

    df_hr = df_hr.drop("returns", axis=1)
    df_hr = df_hr.ffill()
    # Accounts for the application of the minimum number of observations required and
    # merging the two DataFrames. Drop the np.nan values prior to the application of the
    # shift (able to validate the logic).
    df_hr = df_hr.dropna(axis=0, how="any")

    df_hr = df_hr.set_index("real_date", drop=True).shift(1).reset_index(level=0)

    df_hr["cid"] = cross_section

    return df_hr


def adjusted_returns(
    benchmark_return: pd.Series, df_hedge: pd.DataFrame, dfw: pd.DataFrame
) -> pd.DataFrame:
    """
    Method used to compute the hedge ratio returns on the hedging asset which will
    subsequently be subtracted from the returns of the position contracts to calculate
    the adjusted returns (adjusted for the hedged position) across all cross-sections in
    the panel. For instance, if using US Equity to hedge Australia FX:
    AUD_FXXR_NSA_H = AUD_FXXR_NSA - HR_AUD * USD_EQXR_NSA.

    :param <pd.Series> benchmark_return: the return series of the asset being used to
        hedge against the main asset.
    :param <pd.DataFrame> df_hedge: standardised dataframe with the hedge ratios.
    :param <pd.DataFrame> dfw: pivoted dataframe of the relevant returns.

    :return <pd.DataFrame>: standardised dataframe of adjusted returns.
    """

    hedge_pivot = df_hedge.pivot(index="real_date", columns="cid", values="value")
    index = benchmark_return.index

    # Matching the dimensions to the number of assets being hedged.
    benchmark_return = np.tile(
        benchmark_return.to_numpy(), (len(hedge_pivot.columns), 1)
    ).T
    br_df = pd.DataFrame(
        data=benchmark_return, columns=hedge_pivot.columns, index=index
    )

    hedged_returns = hedge_pivot.multiply(br_df)
    adj_rets = dfw - hedged_returns

    df_stack = adj_rets.stack().reset_index(name="value")
    df_stack.columns = ["real_date", "cid", "value"]

    return df_stack


def return_beta(
    df: QuantamentalDataFrame,
    xcat: str = None,
    cids: List[str] = None,
    benchmark_return: str = None,
    start: str = None,
    end: str = None,
    blacklist: dict = None,
    meth: str = "ols",
    oos: bool = True,
    refreq: str = "M",
    min_obs: int = 24,
    max_obs: int = 1000,
    hedged_returns: bool = False,
    ratio_name: str = "_HR",
    hr_name: str = "H",
) -> QuantamentalDataFrame:
    """
    Estimate sensitivities (betas) of return category with respect to single return.

    :param <QuantamentalDataFrame> df: standardized DataFrame with the necessary columns:
        'cid', 'xcat', 'real_date' and 'value.
    :param <str> xcat:  return category based on the type of positions that are
        to be hedged.
        N.B.: Each cross-section of this category uses the same hedge asset/basket.
    :param <List[str]> cids: cross-sections of the returns for which hedge ratios are
        to be calculated. Default is all that are available in the dataframe.
    :param <str> benchmark_return: ticker of return of the hedge asset or basket.
        This is a single series, e.g. U.S. equity index returns ("USD_EQXR_NSA").
    :param <str> start: earliest date in ISO format. Default is None: earliest date in
        df is used.
    :param <str> end: latest date in ISO format. Default is None: latest date in df is
        used.
    :param <dict> blacklist: cross-sections with date ranges that should be excluded from
        the sample of data used for estimating hedge ratios. The estimated ratios
        during blacklist periods will be set equal to the last valid estimate.
    :param <bool> oos: if True (default) hedge ratios are calculated out-of-sample,
        i.e. for the period following the estimation period at the given
        re-estimation frequency.
    :param <str> refreq: re-estimation frequency. This is period after which hedge ratios
        are re-estimated. The re-estimation is conducted at the end of the period and
        used as hedge ratio for all days of the following period. Re-estimation can have
        weekly, monthly, and quarterly frequency with the notations 'W', 'M', and 'Q'
        respectively. The default frequency is monthly.
    :param <int> min_obs: the minimum number of observations required in order to
        estimate a hedge ratio. The default value is 24 days.
        The permissible minimum is 10.
    :param <int> max_obs: the maximum number of latest observations allowed in order to
        estimate a hedge ratio. The default value is 1000.
    :param <str> meth: method used to estimate hedge ratio. At present the only method is
        OLS regression ('ols').
    :param <bool> hedged_returns: If True the function appends the hedged returns to the
        dataframe of hedge ratios. Default is False.
    :param <str> ratio_name: hedge ratio label that will be appended to the category
        name. The default is "_HR". For instance, 'xcat' + "_HR".
    :param <str> hr_name: label used to distinguish the hedged returns in the DataFrame.
        The label is appended to the category being hedged. The default is "H".

    :return <QuantamentalDataFrame>: DataFrame with hedge ratio estimates that update at the
        chosen re-estimation frequency.
        Additionally, the dataframe can include the hedged returns if the parameter
        `benchmark_return` has been set to True.

    N.B.: A return beta is the estimated sensitivity of the main return with respect to
    the asset used for hedging. The ratio is recorded for the period after the estimation
    sample up until the next re-estimation date.

    """

    # Value checks
    if not isinstance(df, pd.DataFrame):
        raise TypeError(f"Expected DataFrame but received <{type(df)}>.")
    df: QuantamentalDataFrame = standardise_dataframe(df)

    all_tix = np.unique(df["cid"] + "_" + df["xcat"])
    bm_error = f"Benchmark return ticker {benchmark_return} is not in the DataFrame."
    if not benchmark_return in all_tix:
        raise ValueError(bm_error)

    error_xcat = (
        f"The field, xcat, must be a string but received <{type(xcat)}>. Only"
        f" a single category is used to hedge against the main asset."
    )
    if not isinstance(xcat, str):
        raise ValueError(error_xcat)

    available_categories = df["xcat"].unique()
    error_hedging = (
        f"The return category used to be hedged, {xcat}, is "
        f"not defined in the dataframe."
    )
    if not xcat in list(available_categories):
        raise ValueError(error_hedging)

    min_obs_error = (
        "The number of minimum observations required to compute a hedge "
        "ratio is 10 business days, or two weeks. Please provide an integer "
        "value greater than 10."
    )
    if not isinstance(min_obs, int) or min_obs < 10:
        raise ValueError(min_obs_error)

    if not isinstance(max_obs, int) or max_obs < min_obs:
        raise ValueError(f"`max_obs` must be an integer ≫ `min_obs`.")

    # Information on hedge return and potential panel adjustment.

    post_fix = benchmark_return.split("_")
    xcat_hedge = "_".join(post_fix[1:])
    cid_hedge = post_fix[0]
    if xcat_hedge == xcat:
        if cid_hedge in cids:
            cids.remove(cid_hedge)
        warnings.warn(
            f"Return to be hedged for cross section {cid_hedge} is the hedge "
            f"return and has been removed from the panel."
        )

    # Wide time series DataFrame of unhedged and benchmark returns.

    # --- Time series DataFrame of unhedged returns.

    dfp = reduce_df(
        df, xcats=[xcat], cids=cids, start=start, end=end, blacklist=blacklist
    )
    dfp_w = dfp.pivot(index="real_date", columns="cid", values="value")
    dfp_w = dfp_w.dropna(axis=0, how="all")

    # --- Time series DataFrame of benchmark return for relevant dates.

    # The asset being used as the hedge could only be defined over a shorter time-period.
    dfh = reduce_df(
        df,
        xcats=[xcat_hedge],
        cids=cid_hedge,
        start=dfp_w.index[0],
        end=dfp_w.index[-1],
    )
    dfh_w = dfh.pivot(index="real_date", columns="cid", values="value")
    dfh_w.columns = ["hedge"]

    # --- Merge time series and calculate re-balancing dates.

    dfw = pd.merge(dfp_w, dfh_w, how="inner", on="real_date")
    br = dfw["hedge"]

    rf = _map_to_business_day_frequency(freq=refreq, valid_freqs=["W", "M", "Q"])
    dates_re = dfw.asfreq(rf).index

    if isinstance(dates_re, pd.DatetimeIndex):
        dates_re: List[str] = dates_re.to_list()

    # Cross-section-wise hedge ratio estimation.

    aggregate = []
    for c in cids:
        df_hr = hedge_calculator(
            unhedged_return=dfw[c],
            benchmark_return=br,
            rdates=dates_re,
            cross_section=c,
            meth=meth,
            min_obs=min_obs,
            max_obs=max_obs,
        )
        aggregate.append(df_hr)

    df_hedge = pd.concat(aggregate, ignore_index=True)

    df_hedge["xcat"] = xcat + ratio_name

    if hedged_returns:
        df_hreturn = adjusted_returns(df_hedge=df_hedge, dfw=dfw, benchmark_return=br)
        df_hreturn = df_hreturn.sort_values(["cid", "real_date"])
        df_hreturn["xcat"] = xcat + "_" + hr_name
        df_hedge = pd.concat([df_hedge, df_hreturn], ignore_index=True)

    return standardise_dataframe(df_hedge)


def beta_display(df_hedge: pd.DataFrame, subplots: bool = False, hr_name: str = "H"):
    """
    Method used to visualise the hedging ratios across the panel: assumes a single
    category is used to hedge the primary asset.

    :param <pd.DataFrame> df_hedge: DataFrame with hedge ratios.
    :param <bool> subplots: matplotlib parameter to determine if each hedging series is
        displayed on separate subplots.
    :param <str> hr_name: label used to distinguish the hedged returns in the DataFrame.
        Comparable to return_beta() method, the default is "H".

    """

    condition = lambda c: c.split("_")[-1] != hr_name

    apply = list(map(condition, df_hedge["xcat"]))
    df_hedge = df_hedge[apply]

    dfw_ratios = df_hedge.pivot(index="real_date", columns="cid", values="value")

    dfw_ratios.plot(subplots=subplots, title="Hedging Ratios.", legend=True)
    plt.xlabel("real_date, years")
    plt.show()


if __name__ == "__main__":
    # Emerging Market Asian countries.
    cids = ["IDR", "INR", "KRW", "MYR", "PHP"]
    # Add the US - used as the hedging asset.
    cids += ["USD"]
    xcats = ["FXXR_NSA", "GROWTHXR_NSA", "INFLXR_NSA", "EQXR_NSA"]

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )

    df_cids.loc["IDR"] = ["2010-01-01", "2020-12-31", 0.5, 2]
    df_cids.loc["INR"] = ["2011-01-01", "2020-11-30", 0, 1]
    df_cids.loc["KRW"] = ["2012-01-01", "2020-11-30", -0.2, 0.5]
    df_cids.loc["MYR"] = ["2013-01-01", "2020-09-30", -0.2, 0.5]
    df_cids.loc["PHP"] = ["2002-01-01", "2020-09-30", -0.1, 2]
    df_cids.loc["USD"] = ["2000-01-01", "2022-03-14", 0, 1.25]

    df_xcats = pd.DataFrame(
        index=xcats,
        columns=["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"],
    )

    df_xcats.loc["FXXR_NSA"] = ["2010-01-01", "2020-10-30", 1, 2, 0.9, 1]
    df_xcats.loc["GROWTHXR_NSA"] = ["2012-01-01", "2020-10-30", 1, 2, 0.9, 1]
    df_xcats.loc["INFLXR_NSA"] = ["2013-01-01", "2020-10-30", 1, 2, 0.8, 0.5]
    df_xcats.loc["EQXR_NSA"] = ["2010-01-01", "2022-03-14", 0.5, 2, 0, 0.2]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    black = {"IDR": ["2010-01-01", "2014-01-04"], "INR": ["2010-01-01", "2013-12-31"]}

    xcat_hedge = "EQXR_NSA"
    # S&P500.
    benchmark_return = "USD_EQXR_NSA"
    df_hedge = return_beta(
        df=dfd,
        xcat=xcat_hedge,
        cids=["IDR", "INR", "KRW", "MYR", "PHP"],
        benchmark_return=benchmark_return,
        start="2010-01-01",
        end="2020-10-30",
        blacklist=black,
        meth="ols",
        oos=True,
        refreq="w",
        min_obs=24,
        hedged_returns=True,
    )

    print(df_hedge)
    beta_display(df_hedge=df_hedge, subplots=False)

    # Long position in S&P500 or the Nasdaq, and subsequently using US FX to hedge the
    # long position.
    xcats = "FXXR_NSA"
    cids = ["USD"]
    benchmark_return = "USD_EQXR_NSA"
    xcat_hedge_two = return_beta(
        df=dfd,
        xcat=xcats,
        cids=cids,
        benchmark_return=benchmark_return,
        start="2010-01-01",
        end="2020-10-30",
        blacklist=black,
        meth="ols",
        oos=True,
        refreq="m",
        min_obs=24,
    )
    print(xcat_hedge_two)



"""
Collection of custom scikit-learn transformer classes.
"""

import numpy as np
import pandas as pd

import datetime

import scipy.stats as stats

from sklearn.linear_model import Lasso, ElasticNet
from sklearn.base import BaseEstimator, TransformerMixin, OneToOneFeatureMixin
from sklearn.feature_selection import SelectorMixin
from sklearn.exceptions import NotFittedError

from statsmodels.tools.tools import add_constant

from linearmodels.panel import RandomEffects

from typing import Union, Any, Optional

import warnings


class ENetSelector(BaseEstimator, SelectorMixin):
    def __init__(
        self, alpha: Union[float, int] = 1.0, l1_ratio=0.5, positive: bool = True
    ):
        """
        Transformer class to use "Elastic Net" as a feature selection algorithm.
        Given two hyper-parameters, `alpha` and `l1_ratio`, the Elastic Net model is fit and
        the non-zero coefficients are used to extract features from an input dataframe.
        If `positive` is set to True, the underlying features as input to the Elastic Net
        are expected to be positively correlated with the target variable.

        :param <Union[float, int]> alpha: the regularisation imposed by the Elastic Net.
        :param <float> l1_ratio: the ratio of L1 to L2 regularisation. If 0, the penalty is
            an L2 penalty. If 1, it is an L1 penalty. If 0 < l1_ratio < 1, the penalty is
            a combination of L1 and L2. As per the `scikit-learn` documentation,
            setting the `l1_ratio` <= 0.01 is not reliable.
        :param <bool> positive: boolean to restrict estimated Elastic Net coefficients to
            be positive.
        """
        if not (isinstance(alpha, float) or isinstance(alpha, int)):
            raise TypeError(
                "The 'alpha' hyper-parameter must be either a float or int."
            )
        if alpha < 0:
            raise ValueError("The 'alpha' hyper-parameter must be non-negative.")

        if not (isinstance(l1_ratio, float) or isinstance(l1_ratio, np.floating)):
            raise TypeError("The 'l1_ratio' hyper-parameter must be a float.")
        if (l1_ratio < 0) or (l1_ratio > 1):
            raise ValueError(
                "The 'l1_ratio' hyper-parameter must be in the interval [0,1]."
            )
        if l1_ratio <= 0.01:
            warnings.warn(
                "Setting the 'l1_ratio' hyper-parameter to be <= 0.01 is not reliable.",
                UserWarning,
            )
        if not isinstance(positive, bool):
            raise TypeError("The 'positive' hyper-parameter must be a boolean.")

        self.alpha = alpha
        self.l1_ratio = l1_ratio
        self.positive = positive
        self.feature_names_in_ = None

    def fit(self, X: pd.DataFrame, y: Union[pd.Series, pd.DataFrame]):
        """
        Fit method to fit an Elastic Net regression and obtain the selected features.

        :param <pd.DataFrame> X: Pandas dataframe of input features.
        :param <Union[pd.Series,pd.DataFrame]> y: Pandas series or dataframe of targets
            associated with each sample in X.
        """
        # checks
        if not isinstance(X, pd.DataFrame):
            raise TypeError(
                "Input feature matrix for the Elastic Net selector must be a pandas dataframe. "
                "If used as part of an sklearn pipeline, ensure that previous steps "
                "return a pandas dataframe."
            )
        if not (isinstance(y, pd.Series) or isinstance(y, pd.DataFrame)):
            raise TypeError(
                "Target vector for the Elastic Net selector must be a pandas series or dataframe. "
                "If used as part of an sklearn pipeline, ensure that previous steps "
                "return a pandas series or dataframe."
            )
        if isinstance(y, pd.DataFrame):
            if y.shape[1] != 1:
                raise ValueError(
                    "The target dataframe must have only one column. If used as part of "
                    "an sklearn pipeline, ensure that previous steps return a pandas "
                    "series or dataframe."
                )
        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("X must be multi-indexed.")
        if not isinstance(y.index, pd.MultiIndex):
            raise ValueError("y must be multi-indexed.")
        if not isinstance(X.index.get_level_values(1)[0], datetime.date):
            raise TypeError("The inner index of X must be datetime.date.")
        if not isinstance(y.index.get_level_values(1)[0], datetime.date):
            raise TypeError("The inner index of y must be datetime.date.")

        if not X.index.equals(y.index):
            raise ValueError(
                "The indices of the input dataframe X and the output dataframe y "
                "don't match as input to the Elastic Net selector."
            )

        self.feature_names_in_ = np.array(X.columns)
        self.p = X.shape[1]

        if self.positive:

            self.enet = ElasticNet(
                alpha=self.alpha, l1_ratio=self.l1_ratio, positive=True
            ).fit(X, y)
        else:
            self.enet = ElasticNet(alpha=self.alpha, l1_ratio=self.l1_ratio).fit(X, y)

        self.selected_ftr_idxs = [i for i in range(self.p) if self.enet.coef_[i] != 0]

        return self

    def _get_support_mask(self):
        """
        Private method to return a boolean mask of the features selected for the Pandas dataframe.
        """
        mask = np.zeros(self.p, dtype=bool)
        mask[self.selected_ftr_idxs] = True
        return mask

    def get_support(self, indices=False):
        """
        Method to return a mask, or integer index, of the features selected for the Pandas dataframe.

        :param <bool> indices: Boolean to specify whether to return the column indices of the selected features instead of a boolean mask

        :return <np.ndarray>: Boolean mask or integer index of the selected features
        """
        if self.feature_names_in_ is None:
            raise NotFittedError(
                "The Elastic Net selector has not been fitted. Please fit the selector before calling get_support()."
            )
        if not isinstance(indices, (bool, np.bool_)):
            raise ValueError("The 'indices' parameter must be a boolean.")
        if indices:
            return self.selected_ftr_idxs
        else:
            mask = self._get_support_mask()
            return mask

    def get_feature_names_out(self):
        """
        Method to mask feature names according to selected features.
        """
        if self.feature_names_in_ is None:
            raise NotFittedError(
                "The Elastic Net selector has not been fitted. Please fit the selector before calling get_feature_names_out()."
            )

        return self.feature_names_in_[self.get_support(indices=False)]

    def transform(self, X: pd.DataFrame):
        """
        Transform method to return only the selected features of the dataframe.

        :param <pd.DataFrame> X: Pandas dataframe of input features.

        :return <pd.DataFrame>: Pandas dataframe of input features selected based
            on the Elastic Net's feature selection capabilities.
        """
        # checks
        if not isinstance(X, pd.DataFrame):
            raise TypeError(
                "Input feature matrix for the Elastic Net selector must be a pandas dataframe. "
                "If used as part of an sklearn pipeline, ensure that previous steps "
                "return a pandas dataframe."
            )
        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("X must be multi-indexed.")
        if not isinstance(X.index.get_level_values(1)[0], datetime.date):
            raise TypeError("The inner index of X must be datetime.date.")
        if not X.shape[-1] == self.p:
            raise ValueError(
                "The number of columns of the dataframe to be transformed, X, doesn't "
                "match the number of columns of the training dataframe."
            )
        if len(self.selected_ftr_idxs) == 0:
            # Then no features were selected
            # Then at the given time, no trading decisions can be made based on these features
            warnings.warn(
                "No features were selected. At the given time, no trading decisions can be made based on these features.",
                UserWarning,
            )
            return X.iloc[:, :0]

        return X.iloc[:, self.selected_ftr_idxs]


class LassoSelector(BaseEstimator, SelectorMixin):
    def __init__(self, alpha: Union[float, int], positive: bool = True):
        """
        Transformer class to use the Lasso as a feature selection algorithm.
        Given a hyper-parameter, alpha, the Lasso model is fit and
        the non-zero coefficients are used to extract features from an input dataframe.
        The underlying features as input to the Lasso model are expected to be positively
        correlated with the target variable.

        :param <Union[float, int]> alpha: the regularisation imposed by the Lasso.
        :param <bool> positive: boolean to restrict estimated Lasso coefficients to
            be positive.
        """
        if not (isinstance(alpha, float) or isinstance(alpha, int)):
            raise TypeError(
                "The 'alpha' hyper-parameter must be either a float or int."
            )
        if alpha < 0:
            raise ValueError("The 'alpha' hyper-parameter must be non-negative.")
        if not isinstance(positive, bool):
            raise TypeError("The 'positive' hyper-parameter must be a boolean.")

        self.alpha = alpha
        self.positive = positive
        self.feature_names_in_ = None

    def fit(self, X: pd.DataFrame, y: Union[pd.Series, pd.DataFrame]):
        """
        Fit method to fit a Lasso regression and obtain the selected features.

        :param <pd.DataFrame> X: Pandas dataframe of input features.
        :param <Union[pd.Series,pd.DataFrame]> y: Pandas series or dataframe of targets
            associated with each sample in X.
        """
        # checks
        if not isinstance(X, pd.DataFrame):
            raise TypeError(
                "Input feature matrix for the LASSO selector must be a pandas dataframe. "
                "If used as part of an sklearn pipeline, ensure that previous steps "
                "return a pandas dataframe."
            )
        if not (isinstance(y, pd.Series) or isinstance(y, pd.DataFrame)):
            raise TypeError(
                "Target vector for the LASSO selector must be a pandas series or dataframe. "
                "If used as part of an sklearn pipeline, ensure that previous steps "
                "return a pandas series or dataframe."
            )
        if isinstance(y, pd.DataFrame):
            if y.shape[1] != 1:
                raise ValueError(
                    "The target dataframe must have only one column. If used as part of "
                    "an sklearn pipeline, ensure that previous steps return a pandas "
                    "series or dataframe."
                )
        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("X must be multi-indexed.")
        if not isinstance(y.index, pd.MultiIndex):
            raise ValueError("y must be multi-indexed.")
        if not isinstance(X.index.get_level_values(1)[0], datetime.date):
            raise TypeError("The inner index of X must be datetime.date.")
        if not isinstance(y.index.get_level_values(1)[0], datetime.date):
            raise TypeError("The inner index of y must be datetime.date.")

        if not X.index.equals(y.index):
            raise ValueError(
                "The indices of the input dataframe X and the output dataframe y "
                "don't match as input to the LASSO selector."
            )

        self.feature_names_in_ = np.array(X.columns)
        self.p = X.shape[1]

        if self.positive:
            self.lasso = Lasso(alpha=self.alpha, positive=True).fit(X, y)
        else:
            self.lasso = Lasso(alpha=self.alpha).fit(X, y)

        self.selected_ftr_idxs = [i for i in range(self.p) if self.lasso.coef_[i] != 0]

        return self

    def _get_support_mask(self):
        """
        Private method to return a boolean mask of the features selected for the Pandas dataframe.
        """
        mask = np.zeros(self.p, dtype=bool)
        mask[self.selected_ftr_idxs] = True
        return mask

    def get_support(self, indices=False):
        """
        Method to return a mask, or integer index, of the features selected for the Pandas dataframe.

        :param <bool> indices: Boolean to specify whether to return the column indices of the selected features instead of a boolean mask

        :return <np.ndarray>: Boolean mask or integer index of the selected features
        """
        if self.feature_names_in_ is None:
            raise NotFittedError(
                "The LASSO selector has not been fitted. Please fit the selector before calling get_support()."
            )
        if not isinstance(indices, (bool, np.bool_)):
            raise ValueError("The 'indices' parameter must be a boolean.")
        if indices:
            return self.selected_ftr_idxs
        else:
            mask = self._get_support_mask()
            return mask

    def get_feature_names_out(self):
        """
        Method to mask feature names according to selected features.
        """
        if self.feature_names_in_ is None:
            raise NotFittedError(
                "The LASSO selector has not been fitted. Please fit the selector before calling get_feature_names_out()."
            )

        return self.feature_names_in_[self.get_support(indices=False)]

    def transform(self, X: pd.DataFrame):
        """
        Transform method to return only the selected features of the dataframe.

        :param <pd.DataFrame> X: Pandas dataframe of input features.

        :return <pd.DataFrame>: Pandas dataframe of input features selected based
            on the Lasso's feature selection capabilities.
        """
        # checks
        if not isinstance(X, pd.DataFrame):
            raise TypeError(
                "Input feature matrix for the LASSO selector must be a pandas dataframe. "
                "If used as part of an sklearn pipeline, ensure that previous steps "
                "return a pandas dataframe."
            )
        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("X must be multi-indexed.")
        if not isinstance(X.index.get_level_values(1)[0], datetime.date):
            raise TypeError("The inner index of X must be datetime.date.")
        if not X.shape[-1] == self.p:
            raise ValueError(
                "The number of columns of the dataframe to be transformed, X, doesn't "
                "match the number of columns of the training dataframe."
            )
        if len(self.selected_ftr_idxs) == 0:
            # Then no features were selected
            # Then at the given time, no trading decisions can be made based on these features
            warnings.warn(
                "No features were selected. At the given time, no trading decisions can be made based on these features.",
                UserWarning,
            )
            return X.iloc[:, :0]

        return X.iloc[:, self.selected_ftr_idxs]


class MapSelector(BaseEstimator, SelectorMixin):
    def __init__(self, threshold: float = 0.05, positive: bool = False):
        """
        Selector class to select features from a training set
        based on the Macrosynergy panel test. This test involves creating
        a linear mixed effects model with period-specific random effects to
        account for cross-sectional correlations. The p-value for the slope
        parameter is used to perform the significance test.

        :param <float> threshold: Significance threshold. This should be in
            the interval (0,1). Default is 0.05.
        :param <bool> positive: Boolean indicating whether or not to only keep features
            with positive estimated model coefficients. Default is False.
        """
        if type(threshold) != float:
            raise TypeError("The threshold must be a float.")
        if (threshold <= 0) or (threshold > 1):
            raise ValueError(
                "The threshold must be in between 0 (inclusive) and 1 (exclusive)."
            )
        if not isinstance(positive, (bool, np.bool_)):
            raise TypeError("The 'positive' parameter must be a boolean.")

        self.threshold = threshold
        self.positive = positive
        self.feature_names_in_ = None

    def fit(self, X: pd.DataFrame, y: Union[pd.Series, pd.DataFrame]):
        """
        Fit method to assess significance of each feature using
        the Macrosynergy panel test.

        :param <pd.DataFrame> X: Pandas dataframe of input features.
        :param <Union[pd.Series, pd.DataFrame]> y: Pandas series or dataframe of targets
            associated with each sample in X.
        """
        # Checks
        if not isinstance(X, pd.DataFrame):
            raise TypeError(
                "Input feature matrix for the MAP selector must be a pandas dataframe. "
                "If used as part of an sklearn pipeline, ensure that previous steps "
                "return a pandas dataframe."
            )
        if not (isinstance(y, pd.Series) or isinstance(y, pd.DataFrame)):
            raise TypeError(
                "Target vector for the MAP selector must be a pandas series or dataframe. "
                "If used as part of an sklearn pipeline, ensure that previous steps "
                "return a pandas series or dataframe."
            )
        if isinstance(y, pd.DataFrame):
            if y.shape[1] != 1:
                raise ValueError(
                    "The target dataframe must have only one column. If used as part of "
                    "an sklearn pipeline, ensure that previous steps return a pandas "
                    "series or dataframe."
                )
        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("X must be multi-indexed.")
        if not isinstance(y.index, pd.MultiIndex):
            raise ValueError("y must be multi-indexed.")
        if not isinstance(X.index.get_level_values(1)[0], datetime.date):
            raise TypeError("The inner index of X must be datetime.date.")
        if not isinstance(y.index.get_level_values(1)[0], datetime.date):
            raise TypeError("The inner index of y must be datetime.date.")
        if not X.index.equals(y.index):
            raise ValueError(
                "The indices of the input dataframe X and the output dataframe y don't "
                "match."
            )

        self.ftrs = []
        self.feature_names_in_ = np.array(X.columns)

        # Convert cross-sections to numeric codes for compatibility with RandomEffects
        unique_xss = sorted(X.index.get_level_values(0).unique())
        xs_codes = dict(zip(unique_xss, range(1, len(unique_xss) + 1)))

        X = X.rename(xs_codes, level=0, inplace=False).copy()
        y = y.rename(xs_codes, level=0, inplace=False).copy()

        # For each column, obtain Wald test p-value
        # Keep significant features
        for col in self.feature_names_in_:
            ftr = X[col]
            ftr = add_constant(ftr)
            # Swap levels so that random effects are placed on each time period,
            # as opposed to the cross-section
            re = RandomEffects(y.swaplevel(), ftr.swaplevel()).fit()
            est = re.params[col]
            zstat = est / re.std_errors[col]
            pval = 2 * (1 - stats.norm.cdf(zstat))
            if pval < self.threshold:
                if self.positive:
                    if est > 0:
                        self.ftrs.append(col)
                else:
                    self.ftrs.append(col)

        return self

    def _get_support_mask(self):
        """
        Private method to return a boolean mask of the features selected for the Pandas dataframe.
        """
        mask = [col in self.ftrs for col in self.feature_names_in_]
        return np.array(mask)

    def get_support(self, indices=False):
        """
        Method to return a mask, or integer index, of the features selected for the Pandas dataframe.

        :param <bool> indices: Boolean to specify whether to return the column indices of the selected features instead of a boolean mask

        :return <np.ndarray>: Boolean mask or integer index of the selected features
        """
        if not isinstance(indices, (bool, np.bool_)):
            raise TypeError("The 'indices' parameter must be a boolean.")
        if self.feature_names_in_ is None:
            raise NotFittedError(
                "The MAP selector has not been fitted. Please fit the selector before calling get_support()."
            )
        mask = self._get_support_mask()

        if indices:
            return np.where(mask)[0]

        return mask

    def get_feature_names_out(self):
        """
        Method to mask feature names according to selected features.
        """
        if self.feature_names_in_ is None:
            raise ValueError(
                "The feature names are not available. Please fit the transformer first."
            )

        return self.feature_names_in_[self.get_support(indices=False)]

    def transform(self, X: pd.DataFrame):
        """
        Transform method to return the significant training features.

        :param <pd.DataFrame> X: Pandas dataframe of input features.

        :return <pd.DataFrame>: Pandas dataframe of input features selected
            based on the Macrosynergy panel test.
        """
        # checks
        if type(X) != pd.DataFrame:
            raise TypeError(
                "Input feature matrix for the MAP selector must be a pandas dataframe. "
                "If used as part of an sklearn pipeline, ensure that previous steps "
                "return a pandas dataframe."
            )
        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("X must be multi-indexed.")
        if not isinstance(X.index.get_level_values(1)[0], datetime.date):
            raise TypeError("The inner index of X must be datetime.date.")
        if not np.all(np.array(X.columns) == self.feature_names_in_):
            raise ValueError(
                "The columns of the dataframe to be transformed, X, don't match the "
                "columns of the training dataframe."
            )
        # transform
        if self.ftrs == []:
            # Then no features were selected
            # Then at the given time, no trading decisions can be made based on these features
            warnings.warn(
                "No features were selected. At the given time, no trading decisions can be made based on these features."
            )
            return X.iloc[:, :0]

        return X[self.ftrs]


class FeatureAverager(BaseEstimator, TransformerMixin):
    def __init__(self, use_signs: Optional[bool] = False):
        """
        Transformer class to combine features into a benchmark signal by averaging.

        :param <Optional[bool]> use_signs: Boolean to specify whether or not to return the
            signs of the benchmark signal instead of the signal itself. Default is False.
        """
        if not isinstance(use_signs, bool):
            raise TypeError("'use_signs' must be a boolean.")

        self.use_signs = use_signs

    def fit(self, X: pd.DataFrame, y: Any = None):
        """
        Fit method. Since this transformer is a simple averaging of features,
        no fitting is required.

        :param <pd.DataFrame> X: Pandas dataframe of input features.
        :param <Any> y: Placeholder for scikit-learn compatibility.
        """

        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Transform method to average features into a benchmark signal.
        If use_signs is True, the signs of the benchmark signal are returned.

        :param <pd.DataFrame> X: Pandas dataframe of input features.

        :return <pd.DataFrame>: Pandas dataframe of benchmark signal.
        """
        # checks
        if type(X) != pd.DataFrame:
            raise TypeError(
                "Input feature matrix for the FeatureAverager must be a pandas dataframe."
                " If used as part of an sklearn pipeline, ensure that previous steps "
                "return a pandas dataframe."
            )
        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("X must be multi-indexed.")
        if not isinstance(X.index.get_level_values(1)[0], datetime.date):
            raise TypeError("The inner index of X must be datetime.date.")

        # transform
        signal_df = X.mean(axis=1).to_frame(name="signal")
        if self.use_signs:
            return np.sign(signal_df).astype(int)

        return signal_df


class ZnScoreAverager(BaseEstimator, TransformerMixin):
    def __init__(self, neutral: str = "zero", use_signs: bool = False):
        """
        Transformer class to combine features into a benchmark signal
        according to a specified normalisation.

        :param <str> neutral: Specified neutral value for a normalisation. This can
            take either 'zero' or 'mean'. If the neutral level is zero, each feature
            is standardised by dividing by the mean absolute deviation. If the neutral
            level is mean, each feature is normalised by subtracting the mean and
            dividing by the standard deviation. Any statistics are computed according
            to a point-in-time principle. Default is 'zero'.
        :param <bool> use_signs: Boolean to specify whether or not to return the
            signs of the benchmark signal instead of the signal itself. Default is False.
        """
        if not isinstance(neutral, str):
            raise TypeError("'neutral' must be a string.")

        if neutral not in ["zero", "mean"]:
            raise ValueError("neutral must be either 'zero' or 'mean'.")

        if not isinstance(use_signs, bool):
            raise TypeError("'use_signs' must be a boolean.")

        self.neutral = neutral
        self.use_signs = use_signs

    def fit(self, X: pd.DataFrame, y: Any = None):
        """
        Fit method to extract relevant standardisation/normalisation statistics from a
        training set so that PiT statistics can be computed in the transform method for
        a hold-out set.

        :param <pd.DataFrame> X: Pandas dataframe of input features.
        :param <Any> y: Placeholder for scikit-learn compatibility.
        """
        # checks
        if not isinstance(X, pd.DataFrame):
            raise TypeError(
                "Input feature matrix for the ZnScoreAverager must be a pandas "
                "dataframe. If used as part of an sklearn pipeline, ensure that previous "
                "steps return a pandas dataframe."
            )
        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("X must be multi-indexed.")
        if not isinstance(X.index.get_level_values(1)[0], datetime.date):
            raise TypeError("The inner index of X must be datetime.date.")

        # fit
        self.training_n: int = len(X)

        if self.neutral == "mean":
            # calculate the mean and sum of squares of each feature
            means: pd.Series = X.mean(axis=0)
            sum_squares: pd.Series = np.sum(np.square(X), axis=0)
            self.training_means: pd.Series = means
            self.training_sum_squares: pd.Series = sum_squares
        else:
            # calculate the mean absolute deviation of each feature
            mads: pd.Series = np.mean(np.abs(X), axis=0)
            self.training_mads: pd.Series = mads

        return self

    def transform(self, X: pd.DataFrame):
        """
        Transform method to compute an out-of-sample benchmark signal for each unique
        date in the input test dataframe. At a given test time, the relevant statistics
        (implied by choice of neutral value) are calculated using all training information
        and test information until (and including) that test time, since the test time
        denotes the time at which the return was available and the features lag behind
        the returns.

        :param <pd.DataFrame> X: Pandas dataframe of input features.
        """
        # checks
        if not isinstance(X, pd.DataFrame):
            raise TypeError(
                "Input feature matrix for the ZnScoreAverager must be a pandas "
                "dataframe. If used as part of an sklearn pipeline, ensure that previous "
                "steps return a pandas dataframe."
            )
        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("X must be multi-indexed.")
        if not isinstance(X.index.get_level_values(1)[0], datetime.date):
            raise TypeError("The inner index of X must be datetime.date.")

        X = X.sort_index(level="real_date")

        # transform
        signal_df = pd.DataFrame(index=X.index, columns=["signal"], dtype="float")

        if self.neutral == "zero":
            training_mads: pd.Series = self.training_mads
            training_n: int = self.training_n
            X_abs = X.abs()

            # We obtain a vector with the number of elements in each expanding window.
            # The count does not necessarily increase uniformly, since some cross-sections
            # may be missing for some dates.
            expanding_count = self._get_expanding_count(X)

            # We divide the expanding sum by the number of elements in each expanding window.
            X_test_expanding_mads = (
                X_abs.groupby(level="real_date").sum().expanding().sum()
                / expanding_count
            )
            n_total = training_n + expanding_count

            X_expanding_mads: pd.DataFrame = (
                (training_n) * training_mads + (expanding_count) * X_test_expanding_mads
            ) / n_total

            standardised_X: pd.DataFrame = (X / X_expanding_mads).fillna(0)

        elif self.neutral == "mean":
            training_means: pd.Series = self.training_means
            training_sum_squares: pd.Series = self.training_sum_squares
            training_n: int = self.training_n

            expanding_count = self._get_expanding_count(X)
            n_total = training_n + expanding_count

            X_square = X**2
            X_test_expanding_sum_squares = (
                X_square.groupby(level="real_date").sum().expanding().sum()
            )
            X_test_expanding_mean = (
                X.groupby(level="real_date").sum().expanding().sum() / expanding_count
            )

            X_expanding_means = (
                (training_n) * training_means
                + (expanding_count) * X_test_expanding_mean
            ) / n_total

            X_expanding_sum_squares = (
                training_sum_squares + X_test_expanding_sum_squares
            )
            comp1 = (X_expanding_sum_squares) / (n_total - 1)
            comp2 = 2 * np.square(X_expanding_means) * (n_total) / (n_total - 1)
            comp3 = (n_total) * np.square(X_expanding_means) / (n_total - 1)
            X_expanding_std: pd.Series = np.sqrt(comp1 - comp2 + comp3)
            standardised_X: pd.DataFrame = (
                (X - X_expanding_means) / X_expanding_std
            ).fillna(0)

        else:
            raise ValueError("neutral must be either 'zero' or 'mean'.")

        benchmark_signal: pd.DataFrame = pd.DataFrame(
            np.mean(standardised_X, axis=1), columns=["signal"], dtype="float"
        )
        signal_df.loc[benchmark_signal.index] = benchmark_signal

        if self.use_signs:
            return np.sign(signal_df).astype(int)

        return signal_df

    def _get_expanding_count(self, X):
        """
        Helper method to get the number of non-NaN values in each expanding window.

        :param <pd.DataFrame> X: Pandas dataframe of input features.

        :return <np.ndarray>: Numpy array of expanding counts.
        """
        return X.groupby(level="real_date").count().expanding().sum().to_numpy()


class PanelMinMaxScaler(BaseEstimator, TransformerMixin, OneToOneFeatureMixin):
    """
    Transformer class to extend scikit-learn's MinMaxScaler() to panel datasets. It is
    intended to replicate the aforementioned class, but critically returning
    a Pandas dataframe or series instead of a numpy array. This preserves the
    multi-indexing in the inputs after transformation, allowing for the passing
    of standardised features into transformers that require cross-sectional
    and temporal knowledge.

    NOTE: This class is designed to replicate scikit-learn's MinMaxScaler() class.
            It should primarily be used to satisfy the assumptions of various models.
    """

    def fit(self, X, y=None):
        """
        Fit method to determine minimum and maximum values over a training set.

        :param <Union[pd.DataFrame, pd.Series]> X: Pandas dataframe or series.
        :param <Any> y: Placeholder for scikit-learn compatibility.

        :return <PanelMinMaxScaler>: Fitted PanelMinMaxScaler object.
        """
        # checks
        if type(X) not in [pd.DataFrame, pd.Series]:
            raise TypeError(
                "'X' must be a pandas dataframe or series. If used as part of an "
                "sklearn pipeline, ensure that previous steps return a pandas dataframe "
                "or series."
            )

        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("X must be multi-indexed.")

        # fit
        self.mins = X.min(axis=0)
        self.maxs = X.max(axis=0)

        return self

    def transform(self, X):
        """
        Transform method to standardise a panel based on the minimum and maximum values.

        :param <Union[pd.DataFrame, pd.Series]> X: Pandas dataframe or series.

        :return <Union[pd.DataFrame, pd.Series]>: Standardised dataframe or series.
        """
        # checks
        if type(X) not in [pd.DataFrame, pd.Series]:
            raise TypeError(
                "'X' must be a pandas dataframe or series. If used as part of an "
                "sklearn pipeline, ensure that previous steps return a pandas dataframe "
                "or series."
            )

        # transform
        calc = (X - self.mins) / (self.maxs - self.mins)

        return calc


class PanelStandardScaler(BaseEstimator, TransformerMixin, OneToOneFeatureMixin):
    def __init__(self, with_mean: bool = True, with_std: bool = True):
        """
        Transformer class to extend scikit-learn's StandardScaler() to panel datasets.
        It is intended to replicate the aforementioned class, but critically returning
        a Pandas dataframe or series instead of a numpy array. This preserves the
        multi-indexing in the inputs after transformation, allowing for the passing
        of standardised features into transformers that require cross-sectional
        and temporal knowledge.

        NOTE: This class is designed to replicate scikit-learn's StandardScaler() class.
                It is not designed to perform sequential mean and standard deviation
                normalisation like the 'make_zn_scores()' function in 'macrosynergy.panel'
                or 'ZnScoreAverager' in 'macrosynergy.learning'.
                This class should primarily be used to satisfy the assumptions of various
                models, for example the Lasso, Ridge or any neural network.

        :param <bool> with_mean: Boolean to specify whether or not to centre the data.
        :param <bool> with_std: Boolean to specify whether or not to scale the data.
        """
        # checks
        if not isinstance(with_mean, bool):
            raise TypeError("'with_mean' must be a boolean.")
        if not isinstance(with_std, bool):
            raise TypeError("'with_std' must be a boolean.")

        # setup
        self.with_mean = with_mean
        self.with_std = with_std

        self.means = None
        self.stds = None

    def fit(self, X: Union[pd.DataFrame, pd.Series], y: Any = None):
        """
        Fit method to determine means and standard deviations over a training set.

        :param <Union[pd.DataFrame, pd.Series]> X: Pandas dataframe or series.
        :param <Any> y: Placeholder for scikit-learn compatibility.

        :return <PanelStandardScaler>: Fitted PanelStandardScaler object.
        """
        # checks
        if not isinstance(X, pd.DataFrame) and not isinstance(X, pd.Series):
            raise TypeError(
                "'X' must be a pandas dataframe or series. If used as part of an sklearn "
                "pipeline, ensure that previous steps return a pandas dataframe or "
                "series."
            )
        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("X must be multi-indexed.")
        # fit
        if self.with_mean:
            self.means = X.mean(axis=0)

        if self.with_std:
            self.stds = X.std(axis=0)

        return self

    def transform(self, X: Union[pd.DataFrame, pd.Series]):
        """
        Transform method to standardise a panel based on the means and standard deviations
        learnt from a training set (and the fit method).

        :param <Union[pd.DataFrame, pd.Series]> X: Pandas dataframe or series.

        :return <Union[pd.DataFrame, pd.Series]>: Standardised dataframe or series.
        """
        # checks
        if type(X) not in [pd.DataFrame, pd.Series]:
            raise TypeError(
                "'X' must be a pandas dataframe or series. If used as part of an sklearn "
                "pipeline, ensure that previous steps return a pandas dataframe or "
                "series."
            )

        # transform
        if self.means is not None:
            calc = X - self.means
        else:
            calc = X

        if self.stds is not None:
            calc = calc / self.stds

        return calc


if __name__ == "__main__":
    from macrosynergy.management import make_qdf
    import macrosynergy.management as msm

    np.random.seed(1)

    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]
    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]

    """Example: Unbalanced panel """

    df_cids2 = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids2.loc["AUD"] = ["2012-01-01", "2020-12-31", 0, 1]
    df_cids2.loc["CAD"] = ["2013-01-01", "2020-12-31", 0, 1]
    df_cids2.loc["GBP"] = ["2010-01-01", "2020-12-31", 0, 1]
    df_cids2.loc["USD"] = ["2010-01-01", "2020-12-31", 0, 1]

    df_xcats2 = pd.DataFrame(index=xcats, columns=cols)
    df_xcats2.loc["XR"] = ["2010-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats2.loc["CRY"] = ["2010-01-01", "2020-12-31", 1, 2, 0.95, 1]
    df_xcats2.loc["GROWTH"] = ["2010-01-01", "2020-12-31", 1, 2, 0.9, 1]
    df_xcats2.loc["INFL"] = ["2010-01-01", "2020-12-31", 1, 2, 0.8, 0.5]

    dfd2 = make_qdf(df_cids2, df_xcats2, back_ar=0.75)
    dfd2["grading"] = np.ones(dfd2.shape[0])
    black = {"GBP": ["2009-01-01", "2012-06-30"], "CAD": ["2018-01-01", "2100-01-01"]}
    dfd2 = msm.reduce_df(df=dfd2, cids=cids, xcats=xcats, blacklist=black)

    dfd2 = dfd2.pivot(index=["cid", "real_date"], columns="xcat", values="value")
    X = dfd2.drop(columns=["XR"])
    y = dfd2["XR"]

    selector = LassoSelector(0.2)
    selector.fit(X, y)
    print(selector.transform(X).columns)

    selector = MapSelector(1e-20)
    selector.fit(X, y)
    print(selector.transform(X).columns)

    # Split X and y into training and test sets
    X_train, X_test = (
        X[X.index.get_level_values(1) < pd.Timestamp(day=1, month=1, year=2018)],
        X[X.index.get_level_values(1) >= pd.Timestamp(day=1, month=1, year=2018)],
    )
    y_train, y_test = (
        y[y.index.get_level_values(1) < pd.Timestamp(day=1, month=1, year=2018)],
        y[y.index.get_level_values(1) >= pd.Timestamp(day=1, month=1, year=2018)],
    )

    selector = ZnScoreAverager(neutral="zero", use_signs=False)
    selector.fit(X_train, y_train)
    print(selector.transform(X_test))

    selector = ZnScoreAverager(neutral="zero")
    selector.fit(X_train, y_train)
    print(selector.transform(X_test))



"""
Functions for visualizing data grading and blacklisted periods from a quantamental DataFrame.
"""
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from typing import List, Tuple

from macrosynergy.management.simulate import make_qdf
from macrosynergy.management.utils import reduce_df


def heatmap_grades(
    df: pd.DataFrame,
    xcats: List[str],
    cids: List[str] = None,
    start: str = "2000-01-01",
    end: str = None,
    grade: str = "grading",
    title: str = None,
    size: Tuple[float] = None,
):
    """Displays a heatmap of grading

    :param <pd.Dataframe> df: standardized DataFrame with the necessary columns:
        'cid', 'xcat', 'real_date' and 'grading'.
    :param List[str] xcats: extended categorys to be checked on.
    :param <List[str]> cids: cross sections to visualize. Default is all in  DataFrame.
    :param <str> start: earliest date in ISO format. Default is earliest available.
    :param <str> end: latest date in ISO format. Default is latest available.
    :param <str> grade: name of column that contains the grades. Default is 'grading'.
    :param <str> title: string of chart title; defaults depend on type of range plot.
    :param <Tuple[float]> size: Tuple of width and height of graph.
        Default is None, meaning it is set in accordance with df.

    """
    df["real_date"] = pd.to_datetime(df["real_date"], format="%Y-%m-%d")

    df_cols = list(df.columns)
    grade_error = (
        "Column that contains the grade values must be present in the "
        f"DataFrame: {df_cols}."
    )
    assert grade in df.columns, grade_error

    df, xcats, cids = reduce_df(df, xcats, cids, start, end, out_all=True)
    df = df[["xcat", "cid", "real_date", grade]]
    df[grade] = df[grade].astype(float).round(2)

    df_ags = (
        df.groupby(["xcat", "cid"])
        .mean()
        .reset_index()
        .pivot(index="xcat", columns="cid", values=grade)
    )

    if size is None:
        size = (max(df_ags.shape[0] / 2, 15), max(1, df_ags.shape[1] / 2))
    if title is None:
        sdate = df["real_date"].min().strftime("%Y-%m-%d")
        title = f"Average grade of vintages since {sdate}"
    sns.set(rc={"figure.figsize": size})
    sns.heatmap(
        df_ags,
        cmap="YlOrBr",
        vmin=1,
        vmax=3,
        annot=True,
        fmt=".1f",
        linewidth=1,
        cbar=False,
    )
    plt.xlabel("")
    plt.ylabel("")
    plt.title(title, fontsize=18)
    plt.show()


if __name__ == "__main__":
    cids = ["NZD", "AUD", "CAD", "GBP"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]
    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD",] = ["2000-01-01", "2020-12-31", 0.1, 1]
    df_cids.loc["CAD",] = ["2001-01-01", "2020-11-30", 0, 1]
    df_cids.loc["GBP",] = ["2002-01-01", "2020-11-30", 0, 2]
    df_cids.loc["NZD",] = ["2002-01-01", "2020-09-30", -0.1, 2]

    df_xcats = pd.DataFrame(
        index=xcats,
        columns=["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"],
    )
    df_xcats.loc["XR",] = ["2000-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["CRY",] = ["2000-01-01", "2020-10-30", 1, 2, 0.95, 1]
    df_xcats.loc["GROWTH",] = ["2001-01-01", "2020-10-30", 1, 2, 0.9, 1]
    df_xcats.loc["INFL",] = ["2001-01-01", "2020-10-30", 1, 2, 0.8, 0.5]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)

    dfd["grading"] = "3"
    filter_date = dfd["real_date"] >= pd.to_datetime("2010-01-01")
    filter_cid = dfd["cid"].isin(["NZD", "AUD"])
    dfd.loc[filter_date & filter_cid, "grading"] = "1"
    filter_date = dfd["real_date"] >= pd.to_datetime("2013-01-01")
    filter_xcat = dfd["xcat"].isin(["CRY", "GROWTH"])
    dfd.loc[filter_date & filter_xcat, "grading"] = "2.1"
    filter_xcat = dfd["xcat"] == "XR"
    dfd.loc[filter_xcat, "grading"] = 1

    heatmap_grades(dfd, xcats=["CRY", "GROWTH", "INFL"], cids=cids)

    dfd.info()



"""
Module for plotting ranges of values across cross-sections for one or more categories.
"""

import pandas as pd
from typing import List, Tuple, Optional

from macrosynergy.management.simulate import make_qdf
import macrosynergy.visuals as msv


def view_ranges(
    df: pd.DataFrame,
    xcats: List[str],
    cids: Optional[List[str]] = None,
    start: str = "2000-01-01",
    end: Optional[str] = None,
    val: str = "value",
    kind: str = "bar",
    sort_cids_by: Optional[str] = None,
    title: Optional[str] = None,
    ylab: Optional[str] = None,
    size: Tuple[float] = (16, 8),
    xcat_labels: Optional[List[str]] = None,
    legend_loc: str = None,
    legend_bbox_to_anchor: Tuple[float] = None,
):
    """Plots averages and various ranges across sections for one or more categories.

    :param <pd.Dataframe> df: standardized DataFrame with the necessary columns:
        'cid', 'xcat', 'real_date' and at least one column with values of interest.
    :param <List[str]> xcats: extended categories to be checked on. Default is all
        in the DataFrame.
    :param <List[str]> cids: cross sections to plot. Default is all in DataFrame.
    :param <str> start: earliest date in ISO format. Default earliest date in df.
    :param <str> end: latest date in ISO format. Default is latest date in df.
    :param <str> val: name of column that contains the values. Default is 'value'.
    :param <str> kind: type of range plot. Default is 'bar'; other option is 'box'.
    :param <str> sort_cids_by: criterion for sorting cids on x-axis;
        Arguments can be 'mean' and 'std'. Default is None, i.e. original order. Ordering
        will be based on the first category if the category is defined over the complete
        panel. Otherwise, mean and standard deviation calculated, of the cross-sections,
        computed across all categories.
    :param <str> title: string of chart title; defaults depend on type of range plot.
    :param <str> ylab: y label. Default is no label.
    :param <Tuple[float]> size: Tuple of width and height of graph. Default is (16, 8).
    :param <List[str]> xcat_labels: custom labels to be used for the ranges.
    :param <str> legend_loc: location of legend; passed to matplotlib.pyplot.legend().
        Default is 'upper center'.
    :param <Tuple[float]> legend_bbox_to_anchor: passed to matplotlib.pyplot.legend().
        Default is (0.5, -0.15).

    """
    if legend_bbox_to_anchor is None and legend_loc is None:
        # -0.15 puts the legend below the x-labels if there are only 2 xcats.
        # If there is more than 2 xcats we need to move the legend further down by a
        # factor of 0.05 for each additional xcat.
        legend_bbox_to_anchor = (0.5, -0.15 - 0.05 * (len(xcats) - 2))
        legend_loc = 'upper center'

    if legend_loc is None:
        legend_loc = 'upper center'
    
    msv.view_ranges(
        df=df,
        xcats=xcats,
        cids=cids,
        start=start,
        end=end,
        val=val,
        kind=kind,
        sort_cids_by=sort_cids_by,
        title=title,
        ylab=ylab,
        size=size,
        xcat_labels=xcat_labels,
        legend_loc=legend_loc,
        legend_bbox_to_anchor=legend_bbox_to_anchor,
    )


if __name__ == "__main__":
    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "CRY"]
    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD",] = ["2010-01-01", "2020-12-31", 0.5, 0.2]
    df_cids.loc["CAD",] = ["2011-01-01", "2020-11-30", 0, 1]
    df_cids.loc["GBP",] = ["2012-01-01", "2020-11-30", 0, 2]
    df_cids.loc["USD",] = ["2012-01-01", "2020-11-30", 1, 2]

    df_xcats = pd.DataFrame(
        index=xcats,
        columns=["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"],
    )
    df_xcats.loc["XR",] = ["2010-01-01", "2020-12-31", 0, 1, 0, 0.3]
    df_xcats.loc["CRY",] = ["2011-01-01", "2020-10-30", 1, 2, 0.9, 0.5]
    df_xcats.loc["INFL",] = ["2011-01-01", "2020-10-30", 1, 2, 0.9, 0.5]
    df_xcats.loc["GROWTH",] = ["2011-01-01", "2020-10-30", 1, 2, 0.9, 0.5]


    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)

    view_ranges(
        dfd,
        xcats=["XR"],
        kind="box",
        start="2012-01-01",
        end="2018-01-01",
        sort_cids_by="std",
    )

    filter_1 = (dfd["xcat"] == "XR") & (dfd["cid"] == "AUD")
    dfd = dfd[~filter_1]

    view_ranges(
        dfd,
        xcats=["XR", "CRY", "INFL"],
        cids=cids,
        kind="box",
        start="2012-01-01",
        end="2018-01-01",
        sort_cids_by=None,
        xcat_labels=["EQXR_NSA", "CRY_NSA", "INFL_NSA"],
    )
